// Author: Vladimir Savic
// E-mail: vladimirsvc@yahoo.com

#include "StateA.h"
#include "StateB.h"
#include "StateC.h"
#include "StateD.h"
#include "StateMachine.h"
#include "TransitionMatrix.h"

// transition matrix:
static BYTE* ptrPreviousState   ;               // pointer to previous state
static BYTE* ptrCurrentState    ;               // pointer to current state
static BYTE  currentStateIndex  ;               // index of currently executed state
static BYTE  prevStateIndex     ;               // index of previous executed state
static BYTE  externalCommand    ;               // external command
static TransitionMatrix* ptrTransitionMatrix;   // pointer to transition matrix

// ***************************************************************************************************************************************************
// *  states encoding (the same encoding as for transition matrix:                                                                                   *
// *  S1 -> 0x00 (first  transition matrix row, index 0) - states[0]                                                                                 *
// *  S2 -> 0x01 (second transition matrix row, index 1) - states[1]                                                                                 *
// *  S3 -> 0x02 (third  transition matrix row, index 2) - states[2]                                                                                 *
// *  S4 -> 0x03 (fourth transition matrix row, index 3) - states[3]                                                                                 *
// ***************************************************************************************************************************************************
static StateAbstract states[NUMBER_OF_STATES];

void Initialization( void* ptrSma ){
    StateMachineAbstract* ptr = (StateMachineAbstract*)ptrSma;
    
    // State machine methods initialization:
    ptr->StateInitialization            = &StateInitialization            ;
    ptr->TransitionMatrixInitialization = &TransitionMatrixInitialization ;
    ptr->ProcessCommand                 = &ProcessCommand                 ;
    ptr->SetCommand                     = &SetCommand                     ;
    ptr->StartProcessing                = &StartProcessing                ;
}

static void StateInitialization( void ){
    states[0].StateInitialization = &StateAInitialization;
    states[1].StateInitialization = &StateBInitialization;
    states[2].StateInitialization = &StateCInitialization;
    states[3].StateInitialization = &StateDInitialization;

    states[0].StateInitialization(&states[0]);
    states[1].StateInitialization(&states[1]);
    states[2].StateInitialization(&states[2]);
    states[3].StateInitialization(&states[3]);
}

void TransitionMatrixInitialization( void* tm ){
    ptrTransitionMatrix = (TransitionMatrix*)tm;
}

void StartProcessing( void ){
    externalCommand  = 0xFF;
    ptrCurrentState  = 0;
    DoTransition(ptrTransitionMatrix->startState);
}

void SetCommand( void* command ){
    BYTE* cmd = (BYTE*)command;
    externalCommand = *cmd;
}

void ProcessCommand ( void ){
    BYTE  offset   = 0 ;
    
    // ptrCurrentState  - pointer to DT byte
    // ptrPreviousState - pointer to PS byte
    // ------------------------------------------------------------------------------
    // |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |
    // ------------------------------------------------------------------------------
    // |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|
    // | Index: 0 | Index: 1 | Index: 2 | Index: 3 | Index: 4 | Index: 5 | Index: 6 |

    // DT: level 1 priority
    if( *ptrCurrentState != 0xFF ) offset = 0;
    // PS: level 2 priority
    else if ( *(ptrCurrentState + 1) == prevStateIndex ) offset = 2;
    // CmdX (external commands): level 3 priority
    else if (*(ptrCurrentState + externalCommand) != 0xFF) offset = externalCommand;
    // interrupt handler (signal/event handler)
    else offset = 0xFF;

    if(offset < NUMBER_OF_TM_COLUMNS)
        DoTransition(ptrCurrentState + offset);
    else
        states[currentStateIndex].InterruptHandler(&externalCommand);
}

static void DoTransition(void* state){
    BYTE* st = (BYTE*)state;
    // ptrCurrentState  - pointer to DT byte
    // ptrPreviousState - pointer to PS byte
    // ------------------------------------------------------------------------------
    // |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |
    // ------------------------------------------------------------------------------
    // |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|
    prevStateIndex = currentStateIndex;
    currentStateIndex = *st;
    ptrCurrentState   = ptrTransitionMatrix->buffer + (NUMBER_OF_TM_COLUMNS * currentStateIndex);   // calculate current state pointer
    states[currentStateIndex].DefaultStateMethod(&externalCommand);                                 // execute default state method
}