// Author: Vladimir Savic
// E-mail: vladimirsvc@yahoo.com

#include "StateMachine.h"
#include "StateAbstract.h"
#include "StateMachineAbstract.h"
#include "TransitionMatrix.h"

// ***************************************************************************************************************************************************
// *  TRANSITION MATRIX STRUCTURE:                                                                                                                   *
// *    <-                        NUMBER_OF_TM_COLUMNS                              ->                                                               *
// *    <-            3 bytes          -><-       external commands 4 bytes         ->                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// *    |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * S1 |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * S2 |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * S3 |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// * S4 |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|                                                               *
// *    ------------------------------------------------------------------------------                                                               *
// *                                                                                                                                                 *
// *  TRANSITION MATRIX LEGEND:                                                                                                                      *
// *  PS   - Previous state                                                                                                                          *
// *  TPS  - Transition by PS  (state machine goes directly to next state only if PS condition is satisfied)                                         *
// *  DT   - Direct transition (when default method is executed, state machine goes directly to another state)                                       *
// *  CmdX - where X belongs to the set of N numbers, present external commands, also called signals/events                                          *
// *  Sx   - where x belongs to the set of N numbers, present state of the state machine                                                             *
// *  0xFF - (255 decimal) means "nothing" (like void in C language) - 0xFF cannot be used for state/cmd encoding                                    *
// *         for example: if DT is set to 0xFF, state does not has direct transition to another state                                                *
// *                      if PS is set to 0xFF, state machine does not look at TPS - transition to another state does not depends on PS              *
// *                                                                                                                                                 *
// *  TRANSITION RULES:                                                                                                                              *
// *  -----------------                                                                                                                              *
// *  Transition types:                                                                                                                              *
// *  PS, TPS  - conditional transition, transition to another state depends on previous state.                                                      *
// *             If previous state is defined TPS field determines next state, all external commands are excluded                                    *
// *  DT       - direct transition, if direct transition is set all external commands are excluded including conditional transition                  *
// *  CmdX     - external command transition. External transitions are included only in case when DT is set to 0xFF as well as PS                    *
// *  Transition priorities (smaller number means higher priority level):                                                                            *
// *  DT   - 1 level                                                                                                                                 *
// *  PS   - 2 level                                                                                                                                 *
// *  CmdX - 3 level                                                                                                                                 *
// *                                                                                                                                                 *
// *  TRANSITION MATRIX FILEDS DEFINITION:                                                                                                           *
// *  each transition matrix filed is determined by 8 bits (BYTE)                                                                                    *
// *                                                                                                                                                 *
// *  ENCODING:                                                                                                                                      *
// *  each state/cmd must be encoded (identified) with 8 bits                                                                                        *
// *                                                                                                                                                 *
// *  states encoding:                                                                                                                               *
// *  S1 -> 0x00 (first  transition matrix row, index 0)                                                                                             *
// *  S2 -> 0x01 (second transition matrix row, index 1)                                                                                             *
// *  S3 -> 0x02 (third  transition matrix row, index 2)                                                                                             *
// *  S4 -> 0x03 (fourth transition matrix row, index 3)                                                                                             *
// *                                                                                                                                                 *
// *  commands encoding (0x00, 0x01 and 0x02 cannot be used for command encoding - those codes are reserved for DT, PS and TPS respectively):        *
// *  encoding for external commands starts from 0x03 ...                                                                                            *
// *  ---------------------------------------------------------------------------------------------------------------------------------------        *
// *  DT   -> always has 0x00 code, transition matrix column index 0                                                                                 *
// *  PS   -> always has 0x01 code, transition matrix column index 1                                                                                 *
// *  TPS  -> always has 0x02 code, transition matrix column index 2                                                                                 *
// *  ---------------------------------------------------------------------------------------------------------------------------------------        *
// *  Cmd1 -> 0x03                                                                                                                                   *
// *  Cmd2 -> 0x04                                                                                                                                   *
// *  Cmd3 -> 0x05                                                                                                                                   *
// *  Cmd4 -> 0x06                                                                                                                                   *
// *                                                                                                                                                 *
// *  SERIAL COMMUNICATION BUFFER STREAM (for this example):                                                                                         *
// *  28 bytes buffer required for transition matrix - TM                                                                                            *
// *  1  byte required for start state identificator - S0                                                                                            *
// *  --------------------------                                                                                                                     *
// *  |    S0    |     TM      |                                                                                                                     *
// *  --------------------------                                                                                                                     *
// *  |<- BYTE ->|<- 28xBYTES ->                                                                                                                     *
// *                                                                                                                                                 *
// *  TM STREAM:                                                                                                                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// *  |    DT    |     PS   |    TPS   |   Cmd1   |   Cmd2   |   Cmd3   |   Cmd4   |           ...             |                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// *  |<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|<- BYTE ->|           ...             |                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// *  |<-                                 S1 State                               ->|<- other states: S2..S4  ->|                                     *
// *  ----------------------------------------------------------------------------------------------------------                                     *
// ***************************************************************************************************************************************************

// Global constants:
const unsigned short BUFFER_SIZE = 29;

// Global variable declaration
BYTE command;
BYTE idata buffer[BUFFER_SIZE]        ;
BYTE bufferIndex                      ;
BYTE portValueState                   ;
BYTE portValueIntHandler              ;
StateMachineAbstract stateMachine     ;
TransitionMatrix     transitionMatrix ;

// ********************
// * Main entry point *
// ********************
void main(){
    UART1_Init( 4800 );                                             // RS 232 communication Initialization
    Delay_ms(1000);
    
    command     = 0xFF ;                                             // set command to 0xFF (for 0xFF details, please look at transition matrix description)
    bufferIndex = 0x00 ;

    portValueState      = 0xFF;
    portValueIntHandler = 0xFF;

    stateMachine.Initialization = &Initialization;                  // state machine Initialization
    stateMachine.Initialization( &stateMachine ) ;
    stateMachine.StateInitialization( )          ;
                                                                    // transition matrix initialization
    transitionMatrix.startState  = &buffer[0];                      // S0  pointer
    transitionMatrix.buffer      = &buffer[1];                      // TM  pointer
    
    stateMachine.TransitionMatrixInitialization(&transitionMatrix); // initialize state machine with transition matrix

    while(1){
        if ( UART1_Data_Ready() == 1 ){
            buffer[bufferIndex++] = UART1_Read();
        }
        else if (bufferIndex == BUFFER_SIZE){
            bufferIndex = 0;
            stateMachine.StartProcessing( );
            command = P3 >> 5;
        }
        else if( command != 0xFF ){
            stateMachine.SetCommand( &command);
            stateMachine.ProcessCommand( );
            command = P3 >> 5;
        }
        P0 = portValueState      ;
        P1 = portValueIntHandler ;
    }
}

void Int0InterruptHandler() org IVT_ADDR_EX0 {
}

void Int1InterruptHandler() org IVT_ADDR_EX1 {
}